#include "clique.h"

int maxInArray(int *array, int n) {
	int i;
	int max = -INF, maxid;
	for(i = 0; i < n; ++ i) {
		if(array[i] > max) {
			max = array[i];
			maxid = i;
		}
	}
	return maxid;
}

void sortArray(int *array, int n) {
	int i, j;
	for(i = 0; i < n; ++ i) {
		for(j = i + 1; j < n; ++ j) {
			if(array[j] < array[i]) {
				/* swap */
				int t;
				t = array[i];
				array[i] = array[j];
				array[j] = t;
			}
		}
	}
}

void clique::findCliques(void){
	/*	the algorithm approximately identifies all non-overlapping cliques	*/
	int i, j, k, m, n;
	int maxNode, maxDegree = -1;
	/*	identifies the maximum degree	*/
	maxNode = maxInArray(cadjCount, numNodes);
	maxDegree = cadjCount[maxNode];

	
	
	/*	check boundary condition	*/
	if(maxDegree < minCliqueSize - 1) {
	/*	note that the node itself counts in the clique	*/
		return;
	} else if (maxDegree == 0) {
	/*	in this case the user want all orphan nodes, output them one by one	*/
		for(i = 0; i < numNodes; ++ i) {
			if(cadjCount[i] == 0) {
				node *n = new node();
				n->count = 1;
				n->contents = new int [n->count];
				n->contents[0] = i;
				addClique(n);
				cadjCount[i] = -1;
			}
		}	
		return;
	}
	/*	construct the possible clique set and the candidate set	*/
	int *possibleClique = new int [maxDegree + 1];
	int *candidates = new int [maxDegree];
	//DEBUG: cout << "after this!" << endl;
	int pIndex = -1, cIndex = -1;
	/*	puts the node with maximum degree into possibleClique set
		and all nodes that connect to it to candidates set	*/
	possibleClique[++ pIndex] = maxNode;
	if(maxDegree > 0) {
		memcpy(candidates, cmatrix[maxNode], maxDegree * sizeof(int));
	}
	cIndex = maxDegree - 1;
	
	/*	repeatedly adding the node in candidates set that has the largest degree */
	while(cIndex >= 0) {
		/* finds the node in candidates set and has the largest degree */
		int localMaxDegree = -1;
		int localMaxNode;
		for(i = 0; i <= cIndex; ++ i) {
			if(cadjCount[candidates[i]] > localMaxDegree) {
				localMaxDegree = cadjCount[candidates[i]];
				localMaxNode = candidates[i];
			}
		}
		if(localMaxDegree < 0) {
			break;		
		}
		/*	puts the node with maximum degree into possibleClique	*/
		/*DEBUG:
		cout << "localMaxDegree:	" << localMaxDegree << endl;
		cout << "localMaxNode:	" << localMaxNode << endl;
		cout << "pIndex:	" << pIndex << endl;
		*/
		possibleClique[++ pIndex] = localMaxNode;
		/*	refine the candidate set, remove the nodes that do not connect to the node just added	*/
		//DEBUG: cout << "maxDegree:	" << maxDegree << endl;
		int *validCandidates = new int [maxDegree];
		//DEBUG: cout << "after this 2" << endl;
		int localVIndex = -1;
		for(i = 0; i <= cIndex; ++ i) {
			/*	as a node is always connect to itself and we should not include such node
				in the next phase, we should skip this node	*/
			if(candidates[i] == localMaxNode) continue;
			if(connected(candidates[i], localMaxNode)) {
				validCandidates[++ localVIndex] = candidates[i];
			}
		}
		if(localVIndex >= 0) {
			memcpy(candidates, validCandidates, (localVIndex + 1) * sizeof(int));
		}
		cIndex = localVIndex;
		delete [] validCandidates;
	}
	
	/*	sort the clique based on node order, for output effect and efficiency of maintaining
		the validity of adjacency matrix after removal of the nodes forming the clique	*/
	sortArray(possibleClique, pIndex + 1);
	
	/* DEBUG
	for(i = 0; i < pIndex; ++ i) {
		if(possibleClique[i] == possibleClique[i + 1]) {
			for(j = 0; j <= pIndex; ++ j) {
				cout << possibleClique[j] << " ";
			}
			cout << endl;
			break;
		}
	}
	*/
	
	/*	output cliques */
	
	if(pIndex + 1 >= minCliqueSize) {
		node *n = new node();
		n->count = pIndex + 1;
		n->contents = new int [n->count];
		for(i = 0; i <= pIndex; ++ i) {
			n->contents[i] = possibleClique[i];
		}
		addClique(n);
	}
	
	/*	maintain the validity of adjacency matrix after removing all nodes
		that have been just identified as a clique	*/
	for(i = 0; i <= pIndex; ++ i) {
		cadjCount[possibleClique[i]] = -1;
	}

	for(i = 0; i < numNodes; ++ i) {
		if(cadjCount[i] > 0) {
		//DEBUG: cout << "checking adjacency matrix of:	" << i << endl;
		/*	reconstruct the adjacency relationship, by removing the nodes that
			just form the clique	*/
			int *newRelation = new int [cadjCount[i]];
			int localNIndex = -1;
			m = n = 0;
			while(m < cadjCount[i] && n <= pIndex) {
				//DEBUG: cout << "comparing:	" << matrix[i][m] << "|" << possibleClique[n] << endl;
				if(cmatrix[i][m] == possibleClique[n]) {
					++ m; ++ n;
				} else if(cmatrix[i][m] > possibleClique[n]) {
					++ n;
				} else if(cmatrix[i][m] < possibleClique[n]) {
					newRelation[++ localNIndex] = cmatrix[i][m];
					++ m;
				}
			}
			/*	if the list of matrix[i] is not exhausted, add all to valid candidate	
				as they will not overlap with the clique formed	*/
			for(j = m; j < cadjCount[i]; ++ j) {
				newRelation[++ localNIndex] = cmatrix[i][j];
			}
			/*	update the matrix array to maintain validity	*/
			if(localNIndex >= 0) {
				memcpy(cmatrix[i], newRelation, (localNIndex + 1) * sizeof(int));
			}
			cadjCount[i] = localNIndex + 1;
			delete [] newRelation;
		}
	}
	
	/*	DEBUG
	for(i = 0; i <= pIndex; ++ i) {
		if(cadjCount[possibleClique[i]] != -1) {
			cout << "Warnning: nodes in clique not removed from itself" << endl;
		}
		for(j = 0; j < numNodes; ++ j) {
			for(k = 0; k < cadjCount[j]; ++ k) {
				if(possibleClique[i] == cmatrix[j][k]) {
					cout << "Warnning: nodes in clique not removed from other nodes" << endl;
				}
			}
		}
	}
	*/		
	delete [] possibleClique;
	delete [] candidates;
	
	
	
	/*	recursively call findCliques	*/
	findCliques();
	return;
}

void clique::addClique(node *n) {
	/* check if it is the first node */
	if(cliqueHead == NULL) {
		cliqueHead = n;
		currentNode = cliqueHead;
	}else {
		currentNode->next = n;
		n->prev = currentNode;
		currentNode = currentNode->next;
	}
}

void clique::mergeNode(void){
	/*	merge the cliques based on the cutoff of clustering coefficient	*/
	/*	backwards order	*/
	int i, j;
	node *p = currentNode;
	if(p->prev == NULL) {
		/*	boundary case	*/
		return;
	}
	node *q = p->prev;
	float cc, maxConnectivity = cCoefficient;
	node *maxNode = NULL;
	/*	search for the best merging site	*/
	while(q != NULL) {
		cc = connectivity(p, q);
		if(cc >= maxConnectivity) {
			maxConnectivity = cc;
			maxNode = q;
		}
		q = q->prev;
	}

	/*	check whether this node should be merged	*/
	if(maxNode != NULL) {
		/*	merge node p to maxNode	*/
		int *newclique = new int [maxNode->count + p->count];
		for(i = 0; i < maxNode->count; ++ i) {
			newclique[i] = maxNode->contents[i];
		}
		for(i = 0; i < p->count; ++ i) {
			newclique[i + maxNode->count] = p->contents[i];
		}
		sortArray(newclique, maxNode->count + p->count);
		
		/*	check for overlapping
		for(i = 0; i < maxNode->count + p->count - 1; ++ i) {
			if(newclique[i] == newclique[i + 1]) {
				cout << "Warnning! Overlapping detected after merging!" << endl;
			}
		}
		*/
		
		delete [] maxNode->contents;
		maxNode->contents = new int [maxNode->count + p->count];
		for(i = 0; i < maxNode->count + p->count; ++ i) {
			maxNode->contents[i] = newclique[i];
		}
		maxNode->count = maxNode->count + p->count;
		delete [] newclique;
		/*	update currentNode and delete node p	*/
		currentNode = currentNode->prev;
		deleteNode(p);
		delete p;
		/*	recursively call mergeNode	*/
		mergeNode();
	} else {
		currentNode = currentNode->prev;
		mergeNode();
	}
	return;
}

void clique::printCliques(void) {
	/* print out the cliques */
	int i;
	node *p = cliqueHead;
	while(p != NULL) {
		for(i = 0; i < p->count; ++ i) {
			cout << p->contents[i] << " ";
		}
		cout << endl;
		p = p->next;
	}
}

