#include "graph.h"
#include "node.h"
#include <stdio.h>
#include <string.h>

#define INF 999999

using namespace std;

class clique : public graph {
	public:
	int minCliqueSize;
	float cCoefficient;
	clique(string filename, int n) : graph(filename, n) {
		cliqueHead = currentNode = NULL;
		/*	make a copy of matrix and adjCount	*/
		cadjCount = new int [numNodes];
		cmatrix = new int* [numNodes];
		int i, j;
		for(i = 0; i < numNodes; ++ i) {
			cadjCount[i] = adjCount[i];
			cmatrix[i] = new int [adjCount[i]];
			for(j = 0; j < adjCount[i]; ++ j) {
				cmatrix[i][j] = matrix[i][j];
			}
		}
	}
	
	~clique(void){
		currentNode = cliqueHead;
		while(currentNode != NULL) {
			node *temp;
			temp = currentNode;
			currentNode = currentNode->next;
			delete temp;
		}
		delete [] cadjCount;
		int i;
		for(i = 0; i < numNodes; ++ i) {
			delete [] cmatrix[i];
		}
		delete [] cmatrix;
	}
	
	void setCutoff(int c) {minCliqueSize = c;}
	void setCoefficient(float e) {cCoefficient = e;}
	void findCliques(void);
	void addClique(node *n);
	void mergeNode(void);
	void printCliques(void);
	
	int cliqueEdges(int n) {
		return (n * (n - 1) / 2);
	}
	
	float connectivity(node *a, node *b) {
		int connect = 0;
		int total = a->count * b->count;
		int i, j;
		for(i = 0; i < a->count; ++ i) {
			for(j = 0; j < b->count; ++ j) {
				if(connected(a->contents[i], b->contents[j])) {
					++ connect;
				}
			}
		}
		return ((float) connect / total);
	}
	
	void addNode(node *toAdd, node *n) {
		//DEBUG: cout << "added new clique" << endl;
		if(toAdd->next == NULL) {
			toAdd->next = n;
			n->prev = toAdd;
		} else {
			toAdd->next->prev = n;
			n->next = toAdd->next;
			toAdd->next = n;
			n->prev = toAdd;
		}
		return;
	}
	
	void deleteNode(node *n) {
		if(n == cliqueHead) {
			cliqueHead = NULL;
			n->prev = n->next = NULL;
			return;
		} else if (n->next == NULL) {
			n->prev->next = NULL;
			n->prev = n->next = NULL;
			return;
		}
		n->prev->next = n->next;
		n->next->prev = n->prev;
		n->prev = n->next = NULL;
		return;
	}
	
	private:
	node *cliqueHead;
	node *currentNode;
	int **cmatrix;
	int *cadjCount;
};
